
import itertools
from collections import OrderedDict

from reahl.component.exceptions import ProgrammerError
from reahl.component.context import ExecutionContext
from reahl.web.fw import PackagedFile, ConcatenatedFile

class LibraryIndex(object):
    def __init__(self, *libraries):
        self.libraries_by_name = OrderedDict()
        for library in libraries:
            self.add(library)

    def clear(self):
        self.libraries_by_name.clear()

    def add(self, library):
        self.libraries_by_name[library.name] = library

    def __contains__(self, name):
        return name in self.libraries_by_name

    def packaged_files(self):
        return [i for i in itertools.chain(*[library.packaged_files() for library in self])]        

    def __iter__(self):
        return iter(self.libraries_by_name.values())
        
    def use_deprecated_yui(self):
        if 'pure' in self:
            del self.libraries_by_name['pure']
        if 'yuigridscss' not in self:
            self.add(YuiGridsCss())
        

class Library(object):
    def __init__(self, name):
        self.name = name
        self.files = []
        self.shipped_in_directory = ''
        self.egg_name = 'reahl-web'
        self.egg_relative_static_root = 'reahl/web/static'

    def packaged_files(self):
        return [PackagedFile(self.egg_name, self.shipped_in_directory, file_name) 
                for file_name in self.files]

    def packaged_files222xxxxxFIXMEIAMDEADButPossibleAlternative(self):
        exposed_files = []
        js_files_to_include = [PackagedFile(self.egg_name, self.shipped_in_directory, file_name) 
                               for file_name in self.files
                               if file_name.endswith('.js')]
        if js_files_to_include:
            exposed_files.append(ConcatenatedFile('%s.js' % self.name, js_files_to_include))

        css_files_to_include = [PackagedFile(self.egg_name, self.shipped_in_directory, file_name) 
                               for file_name in self.files
                               if file_name.endswith('.css')]
        if css_files_to_include:
            exposed_files.append(ConcatenatedFile('%s.css' % self.name, css_files_to_include))
        return exposed_files


    def files_of_type(self, extension):
        return [f for f in self.files if f.endswith(extension)]

    def header_only_material(self, rendered_page):
        result = ''
        for file_name in self.files_of_type('.js'):
            result += '\n<script type="text/javascript" src="/static/%s"></script>' % file_name
        for file_name in self.files_of_type('.css'):
            result += '\n<link rel="stylesheet" href="/static/%s" type="text/css">' % file_name
        return result

    def footer_only_material(self, rendered_page):
        return ''



class JQuery(Library):
    def __init__(self):
        super(JQuery, self).__init__('jquery')
        self.files = ['jquery-1.8.1.js']
        self.shipped_in_directory = '/reahl/web/static/%s' % self.name
        for i in ['jquery.cookie-1.0.js', 
                  'jquery.metadata-2.1.js',
                  'jquery.validate-1.10.0.modified.js',
                  'jquery.ba-bbq-1.2.1.js',
                  'jquery.blockUI-2.43.js',
                  'jquery.form-3.14.js']:
            self.add_shipped_plugin(i)

    def add_shipped_plugin(self, file_name):
        self.files.append(file_name)

    def header_only_material(self, rendered_page):
        return super(JQuery, self).header_only_material(rendered_page) +\
            self.document_ready_material(rendered_page)

    def document_ready_material(self, rendered_page):
        result = '\n<script type="text/javascript">\n'
        result += 'jQuery(document).ready(function($){\n'
        result += '$(\'body\').addClass(\'enhanced\');\n'
        js = set()
        js.update(rendered_page.get_js())
        for item in sorted(js):
            result += item+'\n'
        result += '\n});'
        result += '\n</script>\n'
        return result

    def footer_only_material(self, rendered_page):
        #from http://ryanpricemedia.com/2008/03/19/jquery-broken-in-internet-explorer-put-your-documentready-at-the-bottom/
        return '<!--[if IE 6]>' + self.document_ready_material(rendered_page) +  '<![endif]-->'
        
        

        
class JQueryUI(Library):
    def __init__(self):
        super(JQueryUI, self).__init__('jqueryui')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = ['jquery-ui-1.10.3.custom.js']


class Pure(Library):
    def __init__(self):
        super(Pure, self).__init__('pure')
        self.shipped_in_directory = '/reahl/web/static/pure-release-0.5.0'
        self.files = ['base.css', 'grids.css', 'grids-responsive.css']


class YuiGridsCss(Library):
    @classmethod
    def is_enabled(cls):
        frontend_libraries = ExecutionContext.get_context().config.web.frontend_libraries
        return 'yuigridscss' in frontend_libraries

    @classmethod
    def check_enabled(cls, calling_object):
        if not cls.is_enabled():
            raise ProgrammerError('YuiGridsCss not enabled in current configuration. For %s to work, add the line: \n'\
                                  '"web.frontend_libraries.use_deprecated_yui()"\n in your web.config.py'\
                                   % calling_object.__class__.__name__)

    def __init__(self):
        super(YuiGridsCss, self).__init__('yuigridscss')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = ['reset-fonts-grids.css']


class HTML5Shiv(Library):
    def __init__(self):
        super(HTML5Shiv, self).__init__('html5shiv')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = ['html5shiv-printshiv-3.6.3.js']


class IE9(Library):
    def __init__(self):
        super(IE9, self).__init__('IE9')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = ['IE9.js']

    def header_only_material(self, rendered_page):
        # From: http://remysharp.com/2009/01/07/html5-enabling-script/ 
        result  = '\n<!--[if lt IE 9]>'
        result  += '<script src="/static/html5shiv-printshiv-3.6.3.js" type="text/javascript"></script>'
        result  += '<![endif]-->'
        # From: http://code.google.com/p/ie7-js/ 
        # Not sure if this does not perhaps interfere with Normalize reset stuff? 
        result  += '\n<!--[if lte IE 9]>'  
        result  += '<script src="/static/IE9.js" type="text/javascript"></script>'
        result  += '<![endif]-->'  
        return result



# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import warnings

from nose.tools import istest
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert

from reahl.component.modelinterface import Field
from reahl.web.fw import CannotCreate
from reahl.web.fw import IdentityDictionary
from reahl.web.fw import UrlBoundView
from reahl.web.fw import UserInterface
from reahl.web.fw import Region
from reahl.web.ui import HTML5Page
from reahl.web.ui import P
from reahl.web.pure import PageColumnLayout
from reahl.webdev.tools import Browser
from reahl.webdev.tools import XPath
from reahl.web_dev.fixtures import WebFixture


@istest
class ParameterisedTests(object):
    class ParameterisedScenarios(WebFixture):
        class ParameterisedView(UrlBoundView):
            def assemble(self, some_arg=None):
                if some_arg == 'doesnotexist':
                    raise CannotCreate()
                self.title = 'View for: %s' % some_arg
                self.set_slot('main', P.factory(text='content for %s' % some_arg))

        @scenario
        def normal_arguments(self):
            """Arguments can be sent from where the View is defined."""
            self.argument = 'some arg'
            self.expected_value = 'some arg'
            self.url = '/a_ui/aview'
            self.should_exist = True
        
        @scenario
        def url_arguments(self):
            """Arguments can be parsed from an URL, iff they are specified to the definition as Fields."""
            self.argument = Field()
            self.expected_value = 'test1'
            self.url = '/a_ui/aview/test1'
            self.should_exist = True
            
        @scenario
        def define_entire_page(self):
            """The dynamically defined View can also be define its entire page to be rendered, instead of slots"""
            self.url_arguments() # same setup as this except:
            class SimplePage(HTML5Page):
                def __init__(self, view, some_arg):
                    super(SimplePage, self).__init__(view)
                    self.body.add_child(P(view, text='content for %s' % some_arg))

            class ParameterisedView(UrlBoundView):
                def assemble(self, some_arg=None):
                    if some_arg == 'doesnotexist':
                        raise CannotCreate()
                    self.title = 'View for: %s' % some_arg
                    self.set_page(SimplePage.factory(some_arg))

            self.ParameterisedView = ParameterisedView
            
        @scenario
        def cannot_create(self):
            """To indicate that a view does not exist for the given arguments, the .assemble() 
               method of the View should raise CannotCreate()."""
            self.argument = 'doesnotexist'
            self.url = '/a_ui/aview'
            self.should_exist = False

    @test(ParameterisedScenarios)
    def views_with_parameters(self, fixture):
        """Views can have arguments that originate from code, or are parsed from the URL."""

        class UIWithParameterisedViews(UserInterface):
            def assemble(self):
                self.define_view('/aview', view_class=fixture.ParameterisedView, some_arg=fixture.argument) 

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(PageColumnLayout('main'))
                self.define_user_interface('/a_ui',  UIWithParameterisedViews,  {'main': 'main'}, name='myui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        if fixture.should_exist:
            browser.open(fixture.url)
            vassert( browser.title == 'View for: %s' % fixture.expected_value )
            vassert( browser.is_element_present(XPath.paragraph_containing('content for %s' % fixture.expected_value)) )
        else:
            browser.open(fixture.url, status=404)

    @test(WebFixture)
    def views_from_regex(self, fixture):
        """Parameterised Views can also be added based on a regex over the url."""

        class ParameterisedView(UrlBoundView):
            def assemble(self, some_key=None):
                self.title = 'View for: %s' % some_key
            
        class UIWithParameterisedViews(UserInterface):
            def assemble(self):
                self.define_regex_view('/someurl_(?P<some_key>.*)', '/someurl_${some_key}', view_class=ParameterisedView,
                                       some_key=Field(required=True))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page)
                self.define_user_interface('/a_ui',  UIWithParameterisedViews,  {}, name='myui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # Parameterisedally constructing a View from an URL
        browser.open('/a_ui/someurl_test1')
        vassert( browser.title == 'View for: test1' )

    @test(WebFixture)
    def user_interfaces_from_regex(self, fixture):
        """Sub UserInterfaces can be created on the fly on a UserInterface, based on the URL visited. To indicate that a
           UserInterface does not exist, the creation method should return None."""

        class RegexUserInterface(UserInterface):
            def assemble(self, ui_key=None):
                if ui_key == 'doesnotexist':
                    raise CannotCreate()

                self.name = 'user_interface-%s' % ui_key
                root = self.define_view('/', title='Simple user_interface %s' % self.name)
                root.set_slot('user_interface-slot', P.factory(text='in user_interface slot'))

        class UIWithParameterisedUserInterfaces(UserInterface):
            def assemble(self):
                self.define_regex_user_interface('/apath/(?P<ui_key>[^/]*)',
                                         '/apath/${ui_key}',
                                         RegexUserInterface,
                                         {'user_interface-slot': 'main'},
                                         ui_key=Field())

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(PageColumnLayout('main'))
                self.define_user_interface('/a_ui',  UIWithParameterisedUserInterfaces,  IdentityDictionary(), name='myui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # A sub-user_interface is dynamically created from an URL
        browser.open('/a_ui/apath/test1/')
        vassert( browser.title == 'Simple user_interface user_interface-test1' )

        # The slots of the sub-user_interface is correctly plugged into the page
        [p] = browser.lxml_html.xpath('//p')
        vassert( p.text == 'in user_interface slot' )

        # Another sub-user_interface is dynamically created from an URL
        browser.open('/a_ui/apath/another/')
        vassert( browser.title == 'Simple user_interface user_interface-another' )

        # When the URL cannot be mapped
        browser.open('/a_ui/apath/doesnotexist/', status=404)


    @test(WebFixture)
    def backwards_compatibility(self, fixture):
        """Parameterised, sub UserInterfaces can also be created with define_regex_region, for backwards compatibility.
        """

        class RegexUserInterface(Region):
            def assemble(self, ui_key=None):
                if ui_key == 'doesnotexist':
                    raise CannotCreate()

                self.name = 'user_interface-%s' % ui_key
                root = self.define_view('/', title='Simple user_interface %s' % self.name)
                root.set_slot('user_interface-slot', P.factory(text='in user_interface slot'))

        class UIWithParameterisedUserInterfaces(Region):
            def assemble(self):
                self.define_regex_region('/apath/(?P<ui_key>[^/]*)',
                                         '/apath/${ui_key}',
                                         RegexUserInterface,
                                         {'user_interface-slot': 'main'},
                                         ui_key=Field())

        class MainUI(Region):
            def assemble(self):
                self.define_main_window(HTML5Page).use_layout(PageColumnLayout('main'))
                self.define_region('/a_ui',  UIWithParameterisedUserInterfaces,  IdentityDictionary(), name='myui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # A sub-user_interface is dynamically created from an URL
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter('always')
            browser.open('/a_ui/apath/test1/')
            vassert( browser.title == 'Simple user_interface user_interface-test1' )

        vassert( caught_warnings )


    class ParameterisedUserInterfaceScenarios(WebFixture):
        @scenario
        def normal_arguments(self):
            """Arguments can be sent from where the UserInterface is defined."""
            self.argument = 'some arg'
            self.expected_value = 'some arg'
            self.url = '/a_ui/parameterisedui/aview'
            self.should_exist = True
        
        @scenario
        def url_arguments(self):
            """Arguments can be parsed from an URL, iff they are specified to the definition as Fields."""
            self.argument = Field()
            self.expected_value = 'test1'
            self.url = '/a_ui/parameterisedui/test1/aview'
            self.should_exist = True
            
        @scenario
        def cannot_create(self):
            """To indicate that a UserInterface does not exist for the given arguments, the .assemble() 
               method of the UserInterface should raise CannotCreate()."""
            self.argument = 'doesnotexist'
            self.url = '/a_ui/parameterisedui/aview'
            self.should_exist = False

    @test(ParameterisedUserInterfaceScenarios)
    def parameterised_uis(self, fixture):
        """Sub UserInterfaces can also be parameterised by defining arguments in .define_user_interface, and receiving them in .assemble()."""

        class ParameterisedUserInterface(UserInterface):
            def assemble(self, ui_arg=None):
                if ui_arg == 'doesnotexist':
                    raise CannotCreate()

                self.name = 'user_interface-%s' % ui_arg
                root = self.define_view('/aview', title='Simple user_interface %s' % self.name)
                root.set_slot('user_interface-slot', P.factory(text='in user_interface slot'))


        class UIWithParameterisedUserInterfaces(UserInterface):
            def assemble(self):
                self.define_user_interface('/parameterisedui', ParameterisedUserInterface, {'user_interface-slot': 'main'}, 
                                   ui_arg=fixture.argument,
                                   name='paramui')

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(PageColumnLayout('main'))
                self.define_user_interface('/a_ui',  UIWithParameterisedUserInterfaces,  IdentityDictionary(), name='myui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        if fixture.should_exist:
            browser.open(fixture.url)
            
            # The correct argument was passed
            vassert( browser.title == 'Simple user_interface user_interface-%s' % fixture.expected_value )

            # The slots of the sub-user_interface is correctly plugged into the page
            [p] = browser.lxml_html.xpath('//p')
            vassert( p.text == 'in user_interface slot' )
        else:
            # When the URL cannot be mapped
            browser.open(fixture.url, status=404)








from reahl.sqlalchemysupport import SqlAlchemyControl

reahlsystem.root_egg = 'reahl-web'
reahlsystem.connection_uri = 'postgresql://rhug:rhug@localhost/rhug'
#reahlsystem.connection_uri = 'sqlite:////tmp/test.db'
#reahlsystem.connection_uri = 'sqlite:///:memory:'
reahlsystem.orm_control = SqlAlchemyControl(echo=False)



